#!/usr/bin/python
import os
import sys
import re
import logging
import gzip
import itertools
from config import uri_list, ua_list

delimiter = re.compile('^--------(.*)')

# Create logger for session
def get_logger():
    FORMAT = '%(levelname)s: %(message)s'
    log_config = logging.basicConfig(format=FORMAT, level=logging.INFO)
    logger = logging.getLogger(log_config)
    return logger

# Class to store regex required to parse SIP Log lines
class _RegExLib:
    rsp = re.compile('^SIP\\/2\\.0 ([0-9]{3}).(.*)')
    mtd = re.compile('^CSeq: [0-9]{1,10}.(.*)')
    rq = re.compile('^([A-Z]{3,12}) sips?:.*')
    frm = re.compile('^From: .*<sips?:(?P<FROM_USER>.*)@(.*)')
    to = re.compile('^To: .*<sips?:(?P<TO_USER>.*)@(.*)')
    ua = re.compile('^User-Agent:.(.*)')
    dr = re.compile("^SIP (?P<DIRECTION>\\w{2,3}):")
    rsp_st = re.compile("^SIP\\/2\\.0 ([0-9]{3}).(.*)")
    rqobj = re.compile('^([A-Z]{3,12}).sips?:.*')

    def __init__(self, line):
        self.isresponseparser = self.rsp.match(line)
        self.isrequestparser = self.rq.match(line)
        self.responsemethodparser = self.mtd.match(line)
        self.fromparser = self.frm.match(line)
        self.toparser = self.to.match(line)
        self.uaparser = self.ua.match(line)
        self.direction = self.dr.match(line)
        self.responseStatus = self.rsp_st.match(line)

# Function converts the SIP message into processed objects for further analysis
def create_record(msg_seq, gz_name, logger):

    is_req_msg = False
    is_resp_msg = False
    touser = ""
    fromuser = ""
    direction = ""
    ua = ""
    cseq_methd = ""
    response = ""

    for line in msg_seq:
        reg_match = _RegExLib(line)

        if reg_match.isrequestparser:
            is_req_msg = True
            cseq_methd = reg_match.isrequestparser.group(1)
        if reg_match.isresponseparser:
            is_resp_msg = True
        if reg_match.fromparser:
            fromuser = reg_match.fromparser.group(1)
            
        if reg_match.toparser:
            touser = reg_match.toparser.group(1)
            
        if reg_match.responsemethodparser:
            cseq_methd = reg_match.responsemethodparser.group(1)
            cseq_methd = cseq_methd.strip()

        if reg_match.uaparser:
            ua = reg_match.uaparser.group(1)

        if reg_match.direction:
            if(direction == ""):
                direction = reg_match.direction.group(1)

        if reg_match.responseStatus:
            if(response == ""):
                response = reg_match.responseStatus.group(1)
    
    record = {
            "METHOD": cseq_methd, "FROM_USER": fromuser, 
            "RESPONSE_CODE": response, "DIRECTION": direction, "TO_USER": touser, 
            "USER_AGENT": ua 
        }

    if is_resp_msg and not 100 <= int(response) <= 199:
        if (cseq_methd == "REGISTER"):
            record["OBJECT_TYPE"] = 'SIP_REGISTER_RESPONSE'

        elif (cseq_methd == "INVITE"):
            record["OBJECT_TYPE"] = 'SIP_INVITE_RESPONSE'
        
    elif is_req_msg:
        if (cseq_methd == "REGISTER"):
            record["OBJECT_TYPE"] = 'SIP_REGISTER_REQUEST'

        elif (cseq_methd == "INVITE"):
            record["OBJECT_TYPE"] = 'SIP_INVITE_REQUEST'
    else:
        logger.debug("sequence with unidentified method")

    return record

# Function to run filter for Number of Register rejects
def reg_reject_cond(record):
    reg_rej = ["401","200"]
    if record["METHOD"] == "REGISTER" and record["RESPONSE_CODE"] and record["TO_USER"]:
        if record["RESPONSE_CODE"] not in reg_rej:
            return True
        else:
            return False 
    else:
        return False 

# Function to run filter for Number of Invite rejects
def inv_reject_cond(record):
    invite_rej = ["407", "200"]
    if record["METHOD"] == "INVITE" and record["RESPONSE_CODE"] and record["FROM_USER"] and record["DIRECTION"] == "IN":
        if record["RESPONSE_CODE"] not in invite_rej:
            return True
        else:
            return False
    else:
        return False

# Function to run filter for Number of unexpected URI outside your range of extension
def uri_cond(record):
    if record["METHOD"] == "REGISTER" and record["FROM_USER"]:
        if any(record["FROM_USER"].startswith(str(uri)) for uri in uri_list):
            return False
        else:
            return True
    else:
        return False

# Function to run filter for Number of unexpected user-agents outside your ecosystem of devices
def ua_cond(record):
    if record["METHOD"] == "REGISTER" and record["USER_AGENT"]:
        if any(record["USER_AGENT"].startswith(str(ua)) for ua in ua_list):
            return False
        else:
            return True
    else:
        return False

# Function to run filter for Register rejects and returns To User
def get_reject_ext_reg(record):
    reg_rej = ["401","200"]
    if record["METHOD"] == "REGISTER" and record["RESPONSE_CODE"] and record["TO_USER"]:
        if record["RESPONSE_CODE"] not in reg_rej:
            return record["TO_USER"]  
        else:
            return False
    else:
        False

# Function to run filter for Register rejects and returns From User
def get_reject_num_inv(record):
    invite_rej = ["407", "200"]
    if record["METHOD"] == "INVITE" and record["RESPONSE_CODE"] and record["FROM_USER"] and record["DIRECTION"] == "IN":
        if record["RESPONSE_CODE"] not in invite_rej:
            return record["FROM_USER"] + "-" + record["DIRECTION"]
        else:
            return False
    else:
        return False

# Function to run filter for unexpected user-agents and returns User Agent
def get_susp_ua(record):
    if record["METHOD"] == "REGISTER" and record["USER_AGENT"]:
        if any(record["USER_AGENT"].startswith(str(ua)) for ua in ua_list):
            return False
        else:
            return record["USER_AGENT"].strip()
    else:
        return False

# Function to run filter for unexpected URI and returns From User
def get_unexpected_uri(record):
    if record["METHOD"] == "REGISTER" and record["FROM_USER"]:
        if any(record["FROM_USER"].startswith(str(uri)) for uri in uri_list):
            return False
        else:
            return record["FROM_USER"]
    else:
        return False

# Generic Function to find top five entries of input 
def get_top_5_general(main_dict, new_dict):
    try:
        for ext_or_ua, count in new_dict.items():
            if ext_or_ua in main_dict:
                main_dict[ext_or_ua] += count
            else:
                main_dict[ext_or_ua] = count

        top_ua = dict(sorted(main_dict.items(), key = lambda x: x[1], reverse = True))
        ret_dict = dict(itertools.islice(top_ua.items(), 5))
        return ret_dict
    except Exception as e:
        print("exception while sorting top 5 ext/ua/uri ",e)
        return main_dict

# Function to display top registration rejects on terminal
def display_top_reg_rej(top_res):
    print(" "+"-"*43)
    print(" |{:^20}|{:^20}|".format("Extension", "Rejects"))
    print(" "+"-"*43)
    for ext, count in itertools.islice(top_res.items(), 5):
        print(" |{:^20}|{:>20}|".format(ext, count))
    print(" "+"-"*43)

# Function to display top invite rejects on terminal
def display_top_inv_rej(top_res):
    print(" "+"-"*54)
    print(" |{:^20}|{:^20}|{:^10}|".format("Caller", "Rejects", "DIR"))
    print(" "+"-"*54)
    for ext, count in itertools.islice(top_res.items(), 5):
        num, inv_dir = ext.split("-")
        print(" |{:^20}|{:>20}|{:^10}|".format(num, count, inv_dir))
    print(" "+"-"*54)

# Function to display top unexpected User Agents on terminal
def display_top_unkn_ua(top_res):
    print(" "+"-"*64)
    print(" |{:^42}|{:^19}|".format("Unexpected UA", "Register attempts"))
    print(" "+"-"*64)
    for ua, count in itertools.islice(top_res.items(), 5):
        ua = ua.strip() if len(ua.strip()) <= 40 else ua[:40].strip()
        print(" |{:^42}|{:>19}|".format(ua, count))
    print(" "+"-"*64)

# Function to display top unexpected URI on terminal
def display_top_uri(top_res):
    print(" "+"-"*43)
    print(" |{:^20}|{:^20}|".format("Unexpected URI", "Register"))
    print(" "+"-"*43)
    for uri, count in itertools.islice(top_res.items(), 5):
        print(" |{:^20}|{:>20}|".format(uri, count))
    print(" "+"-"*43)

def exit_script():
    input("\nPress any key to exit.\n")
    sys.exit(0)

# The function does the following
# 1. Get the path of the directory where the trace gzip files are location
# 2. Parses the SIP messages from this file (INVITE and REGISTER only)
# 3. Evaluates the failed REGISTER and INVITE messages
# 4. Prints the evaluated results to indicate if there is potential attack observed
def main():
    try:
        logger = get_logger()
        gzip_path = input("Enter the path of your gzip directory (absolute path): ")
        print("\n Processing log files placed at:",gzip_path)

        if os.path.isdir(gzip_path):
            if os.listdir(gzip_path):
                pass
            else:
                logger.info("Directory is empty: {}".format(gzip_path))
                return None
        else:
            logger.info("Directory not exists: {}".format(gzip_path))
            return None

        num_of_files = 0
        num_of_lines = 0
        recs_count = 0
        reg_reject_count = 0
        inv_reject_count = 0
        unkn_uri_count = 0
        unkn_ua_count = 0
        top5_ext_n_count_reg_rej = {}
        top5_num_n_count_inv_rej = {}
        top5_susp_ua_n_count = {}
        top5_unexcp_uri_n_count = {}
        

        with os.scandir(gzip_path) as it:
            for entry in it:
                
                if entry.name.endswith(".gz") and entry.is_file():

                    gz_name = entry.name
                    num_of_files += 1

                    with gzip.open(entry, mode="rb") as gzip_file:
                        sip_seq = []
                        batch_to_analyse = []
                        target_ext_count_reg = {}
                        target_ext_count_inv = {}
                        target_ua_count_map = {}
                        target_uri_count_map = {}

                        for line in gzip_file:
                            num_of_lines +=1
                            try:
                                decoded_line = line.decode('utf-8')
                            except Exception as e:
                                decoded_line = line.decode('ISO-8859-1')

                            if delimiter.match(decoded_line):

                                sip_record = create_record(sip_seq, gz_name, logger)
                                batch_to_analyse.append(sip_record)
                                sip_seq.clear()
                                recs_count +=1

                                try:
                                    is_rej_ext = get_reject_ext_reg(sip_record)
                                    if is_rej_ext:
                                        if is_rej_ext in target_ext_count_reg:
                                            target_ext_count_reg[is_rej_ext] += 1
                                        else:
                                            target_ext_count_reg[is_rej_ext] = 1
                                except Exception as e:
                                    logger.debug("exception while counting top register rejects: filename: {}".format(gz_name))
                                
                                try:
                                    is_rej_ext = get_reject_num_inv(sip_record)
                                    if is_rej_ext:
                                        if is_rej_ext in target_ext_count_inv:
                                            target_ext_count_inv[is_rej_ext] += 1
                                        else:
                                            target_ext_count_inv[is_rej_ext] = 1
                                except Exception as e:
                                    logger.debug("exception while counting top register rejects: filename: {}".format(gz_name))
                                
                                try:
                                    is_ua_susp = get_susp_ua(sip_record)
                                    if is_ua_susp:
                                        if is_ua_susp in target_ua_count_map:
                                            target_ua_count_map[is_ua_susp] += 1
                                        else:
                                            target_ua_count_map[is_ua_susp] = 1
                                except Exception as e:
                                    logger.debug("exception while counting top suspicious user agent registrations: filename: {}".format(gz_name))

                                try:
                                    is_uri_susp = get_unexpected_uri(sip_record)
                                    if is_uri_susp:
                                        if is_uri_susp in target_uri_count_map:
                                            target_uri_count_map[is_uri_susp] += 1
                                        else:
                                            target_uri_count_map[is_uri_susp] = 1

                                except Exception as e:
                                    logger.debug("exception while counting top suspicious URI rejects registrations: filename: {}".format(gz_name))

                            else:
                                sip_seq.append(decoded_line.strip())

                        try:
                            reg_reject_count += sum(map(reg_reject_cond, batch_to_analyse))
                        except Exception as e:
                            logger.exception("exception while counting Register Reject: filename: {}".format(gz_name))
                        
                        try:
                            inv_reject_count += sum(map(inv_reject_cond, batch_to_analyse))
                        except Exception as e:
                            logger.exception("exception while counting Invite Reject: filename: {}".format(gz_name))
                        
                        try:
                            unkn_ua_count += sum(map(ua_cond, batch_to_analyse))
                        except Exception as e:
                            logger.debug("exception while checking unknown User Agent: filename: {}".format(gz_name))

                        try:
                            unkn_uri_count += sum(map(uri_cond, batch_to_analyse))
                        except Exception as e:
                            logger.debug("exception while checking unknown URI: filename: {}".format(gz_name))
                                
                    if target_ext_count_reg:
                        sorted_ext_reg = dict(sorted(target_ext_count_reg.items(), key = lambda x: x[1], reverse = True))
                        top_5_of_batch = dict(itertools.islice(sorted_ext_reg.items(), 5))
                        top5_ext_n_count_reg_rej = get_top_5_general(top5_ext_n_count_reg_rej, top_5_of_batch)
                    
                    if target_ext_count_inv:
                        sorted_ext_inv = dict(sorted(target_ext_count_inv.items(), key = lambda x: x[1], reverse = True))
                        top_5_of_batch = dict(itertools.islice(sorted_ext_inv.items(), 5))
                        top5_num_n_count_inv_rej = get_top_5_general(top5_num_n_count_inv_rej, top_5_of_batch)

                    if target_ua_count_map:
                        sorted_susp_ua = dict(sorted(target_ua_count_map.items(), key = lambda x: x[1], reverse = True))
                        top_5_of_batch = dict(itertools.islice(sorted_susp_ua.items(), 5))
                        top5_susp_ua_n_count = get_top_5_general(top5_susp_ua_n_count, top_5_of_batch)

                    if target_uri_count_map:
                        sorted_unexcp_uri = dict(sorted(target_uri_count_map.items(), key = lambda x: x[1], reverse = True))
                        top_5_of_batch = dict(itertools.islice(sorted_unexcp_uri.items(), 5))
                        top5_unexcp_uri_n_count = get_top_5_general(top5_unexcp_uri_n_count, top_5_of_batch)

        if reg_reject_count > 0 or inv_reject_count > 0 or unkn_ua_count > 0 or unkn_uri_count > 0:
            print()
            print(" ****************************************************")
            print(" *                                                  *")
            print(" *       Your SBC is under a potential attack       *")
            print(" *                                                  *")
            print(" ****************************************************\n")
        
        # Total Register Rejects
        print("\n 1. Number of register rejects:",reg_reject_count)

        # Top Register Rejects
        if len(top5_ext_n_count_reg_rej) >= 5:
            print(" Top 5 extensions targeted")
            display_top_reg_rej(top5_ext_n_count_reg_rej)
        elif len(top5_ext_n_count_reg_rej) > 0:
            print(" Top extensions targeted")
            display_top_reg_rej(top5_ext_n_count_reg_rej)

        print("\n "+"-"*100)
        # Total Invite Rejects
        print("\n 2. Number of invite rejects:",inv_reject_count)

        # Top Invite Rejects
        if len(top5_num_n_count_inv_rej) >= 5:
            print(" Top 5 callers rejected")
            display_top_inv_rej(top5_num_n_count_inv_rej)
        elif len(top5_num_n_count_inv_rej) > 0:
            print(" Top 5 callers rejected")
            display_top_inv_rej(top5_num_n_count_inv_rej)

        print("\n "+"-"*100)
        # Total Unexpected user-agents
        print("\n 3. Number of unexpected user-agents outside your ecosystem of devices:",unkn_ua_count)
        print(" Expected User-agents starting with: {}".format(', '.join(map(str, ua_list))))

        # Top Unexpected user-agents
        if len(top5_susp_ua_n_count) >= 5:
            print("\n Top 5 unexpected user-agents registration attempts")
            display_top_unkn_ua(top5_susp_ua_n_count)
        elif len(top5_susp_ua_n_count) > 0:
            print("\n Top unexpected user-agents registration attempts")
            display_top_unkn_ua(top5_susp_ua_n_count)

        print("\n "+"-"*100)
        # Total Unexpected URI
        print("\n 4. Number of unexpected URI outside your range of extension:",unkn_uri_count)
        print(" Expected URI starting with: {}".format(', '.join(map(str, uri_list))))

        # Top Unexpected URI 
        if len(top5_unexcp_uri_n_count) >= 5:
            print("\n Top 5 unexpected URI registration attempts")
            display_top_uri(top5_unexcp_uri_n_count)
        elif len(top5_unexcp_uri_n_count) > 0:
            print("\n Top unexpected URI registration attempts")
            display_top_uri(top5_unexcp_uri_n_count)

        # general processing stats
        print("\n Processed:",num_of_lines,"log lines from",num_of_files,"log files")

        if reg_reject_count > 0 or inv_reject_count > 0 or unkn_ua_count > 0 or unkn_uri_count > 0:
            print("\n To get actionable intelligence, scan your SBC using Assertion https://scan.assertion.cloud.")
        else:
            print("\n We did not find evidence of attack on your SBC")
    except Exception as e:
        print("unknown error occurred while performing operations")
        print("error message:",e)

    
if __name__ == "__main__":
    main()
    exit_script()